import random
import inspect
import sys

from deap import base, creator, tools, gp, algorithms
import numpy

import behavior_tree as bt

# primitives for the inner nodes
def selector_1_child(child1):
    return bt.SelectorNode([child1])


def selector_2_children(child1, child2):
    return bt.SelectorNode([child1, child2])


def selector_3_children(child1, child2, child3):
    return bt.SelectorNode([child1, child2, child3])


def selector_star_1_child(child1):
    return bt.SelectorStarNode([child1])


def selector_star_2_children(child1, child2):
    return bt.SelectorStarNode([child1, child2])


def selector_star_3_children(child1, child2, child3):
    return bt.SelectorStarNode([child1, child2, child3])


def sequence_1_child(child1):
    return bt.SequenceNode([child1])


def sequence_2_children(child1, child2):
    return bt.SequenceNode([child1, child2])


def sequence_3_children(child1, child2, child3):
    return bt.SequenceNode([child1, child2, child3])


def sequence_star_1_child(child1):
    return bt.SequenceStarNode([child1])


def sequence_star_2_children(child1, child2):
    return bt.SequenceStarNode([child1, child2])


def sequence_star_3_children(child1, child2, child3):
    return bt.SequenceStarNode([child1, child2, child3])


# primitives for the leaf nodes
def condition_black_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(0, params)


def condition_gray_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(1, params)


def condition_white_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(2, params)


def condition_neighborcount(eta, xi):
    params = {"w": eta, "p": xi}
    return bt.ConditionNode(3, params)


def condition_inverted_neighborcount(eta, xi):
    params = {"w": eta, "p": xi}
    return bt.ConditionNode(4, params)


def condition_fixed_probability(beta):
    params = {"p": beta}
    return bt.ConditionNode(5, params)


def condition_receiving_signal(signal):
    params = {"sig": signal}
    return bt.ConditionNode(9, params)


class ParamBeta(float):
    pass


class ParamXi(int):
    pass


class ParamEta(float):
    pass


class ParamSignal(int):
    pass


class ParamSignalStar(int):
    pass


def action_exploration(rwm):
    params = {"rwm": rwm}
    return bt.ActionNode(0, params)


def action_stop():
    params = {}
    return bt.ActionNode(1, params)


def action_grouping(n_min, n_max, att):
    params = {"min": n_min, "max": n_max, "att": att}
    return bt.ActionNode(9, params)


def action_isolation(n_min, n_max, rep):
    params = {"min": n_min, "max": n_max, "rep": rep}
    return bt.ActionNode(10, params)


def action_meeting(sig, dmin, att):
    params = {"sig": sig, "dmin": dmin, "att": att}
    return bt.ActionNode(12, params)


def action_acknowledgement(sig, tmax):
    params = {"sig": sig, "tmax": tmax}
    return bt.ActionNode(13, params)


def action_emit_signal(sig):
    params = {"sig": sig}
    return bt.ActionNode(11, params)


class ParamNMax(int):
    pass


class ParamNMin(int):
    pass


class ParamDMin(float):
    pass


class ParamTMax(int):
    pass


class ParamSigZero(int):
    pass


class ParamAttRep(float):
    pass

class ParamRWM(int):
    pass


def create_primitive_set():
    # defined a new primitive set for strongly typed GP
    pset = gp.PrimitiveSetTyped("MAIN", [], bt.ControlFlowNode)

    # operator for selector node
    pset.addPrimitive(selector_1_child, [bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(selector_2_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(selector_3_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)

    # operator for selector* node
    pset.addPrimitive(selector_star_1_child, [bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(selector_star_2_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(selector_star_3_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)

    # operator for sequence node
    pset.addPrimitive(sequence_1_child, [bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(sequence_2_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(sequence_3_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)

    # operator for sequence* node
    pset.addPrimitive(sequence_star_1_child, [bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(sequence_star_2_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)
    pset.addPrimitive(sequence_star_3_children, [bt.BehaviorTreeNode, bt.BehaviorTreeNode, bt.BehaviorTreeNode], bt.ControlFlowNode)

    # operator for condition node
    pset.addPrimitive(condition_black_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_gray_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_white_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_neighborcount, [ParamEta, ParamXi], bt.ConditionNode)
    pset.addPrimitive(condition_inverted_neighborcount, [ParamEta, ParamXi], bt.ConditionNode)
    pset.addPrimitive(condition_fixed_probability, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_receiving_signal, [ParamSignalStar], bt.ConditionNode)

    # operator for action node
    pset.addPrimitive(action_exploration, [ParamRWM], bt.ActionNode)
    pset.addPrimitive(action_stop, [], bt.ActionNode)
    pset.addPrimitive(action_grouping, [ParamNMin, ParamNMax, ParamAttRep], bt.ActionNode)
    pset.addPrimitive(action_isolation, [ParamNMin, ParamNMax, ParamAttRep], bt.ActionNode)
    pset.addPrimitive(action_meeting, [ParamSignalStar, ParamDMin, ParamAttRep], bt.ActionNode)
    pset.addPrimitive(action_acknowledgement, [ParamSignal, ParamTMax], bt.ActionNode)
    pset.addPrimitive(action_emit_signal, [ParamSignal], bt.ActionNode)

    # terminals
    pset.addEphemeralConstant("param_beta", lambda: random.uniform(0, 1), ParamBeta)
    pset.addEphemeralConstant("param_eta", lambda: random.uniform(0, 20), ParamEta)
    pset.addEphemeralConstant("param_xi", lambda: random.randint(0, 10), ParamXi)
    pset.addEphemeralConstant("param_signal", lambda: random.randint(1, 6), ParamSignal)
    pset.addEphemeralConstant("param_signal_star", lambda: random.randint(1, 7), ParamSignalStar)
    pset.addEphemeralConstant("param_rwm", lambda: random.randint(0, 100), ParamRWM)
    pset.addEphemeralConstant("param_att_rep", lambda: random.uniform(1, 5), ParamAttRep)
    pset.addEphemeralConstant("param_signal_zero", lambda: random.randint(0, 6), ParamSigZero)
    pset.addEphemeralConstant("param_tmax", lambda: random.randint(0, 20), ParamTMax)
    pset.addEphemeralConstant("param_dmin", lambda: random.uniform(0.1, 1.0), ParamDMin)
    pset.addEphemeralConstant("param_n_min", lambda: random.randint(0, 20), ParamNMin)
    pset.addEphemeralConstant("param_n_max", lambda: random.randint(0, 20), ParamNMax)

    return pset

def choose_control_flow_node_primitive(pset):
    possible_primitives = pset.primitives[bt.ControlFlowNode]  # pset.primitives[bt.SelectorNode] + pset.primitives[bt.SelectorStarNode] + pset.primitives[bt.SequenceNode] + pset.primitives[bt.SequenceStarNode]
    primitive = random.choice(possible_primitives)
    return primitive

def choose_leaf_node_primitive(pset):
    possible_primitives = pset.primitives[bt.ActionNode] + pset.primitives[bt.ConditionNode]
    primitive = random.choice(possible_primitives)
    return primitive

def generate_bt(pset, min_, max_, depth=0):
    expr = []
    if depth == max_:
        # choose leaf node
        prim = choose_leaf_node_primitive(pset)
        expr.append(prim)
        for arg in prim.args:
            term = random.choice(pset.terminals[arg])()
            expr.append(term)
    elif depth < min_:
        # choose a control flow node
        prim = choose_control_flow_node_primitive(pset)
        expr.append(prim)
        for arg in prim.args:
            expr.extend(generate_bt(pset, min_, max_, depth+1))
    else:
        # choose a control flow node or a leaf node
        is_leaf_node = random.choice([True, False])
        if is_leaf_node:
            prim = choose_leaf_node_primitive(pset)
            expr.append(prim)
            for arg in prim.args:
                term = random.choice(pset.terminals[arg])()
                expr.append(term)
        else:
            prim = choose_control_flow_node_primitive(pset)
            expr.append(prim)
            for arg in prim.args:
                expr.extend(generate_bt(pset, min_, max_, depth+1))
    return expr

def create_toolbox(pset):
    toolbox = base.Toolbox()
    # toolbox.register("expr", gp.genFull, pset=pset, min_=1, max_=10)
    toolbox.register("expr", generate_bt, pset=pset, min_=1, max_=2)  # depth 0 is the toplevel node
    toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("compile", gp.compile, pset=pset)

    toolbox.register("evaluate", , toolbox=toolbox)

    return toolbox


def create_hall_of_fame():
    hof = tools.HallOfFame(1)
    return hof


def create_stats():
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", numpy.mean)
    stats.register("std", numpy.std)
    stats.register("min", numpy.min)
    stats.register("max", numpy.max)
    return stats


def setup_gp():
    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax)

    pset = create_primitive_set()
    toolbox = create_toolbox(pset)
    hof = create_hall_of_fame()
    stats = create_stats()

    return pset, toolbox, hof, stats


def print_individual(individual):
    print(individual)
    for node in individual:
        if not isinstance(node, gp.Terminal):
            print("    " * compute_depth_of_primitive(individual, node) + "{} ({}): {}".format(node.name, node.arity, node.args))
            # print("    {} ({}): {}".format(node.name, node.arity, node.args))


def compute_depth_in_subtree(individual, primitive, index):
    if index >= len(individual):
        print("Index outside of range")
        return None
    # print("Comparing {} and {}".format(individual[index], primitive))
    if individual[index] == primitive:
        return 0
    child_index = index + 1
    for i in range(len(individual[index].args)):
        child_subtree_slice = individual.searchSubtree(child_index)
        if primitive in individual[child_subtree_slice]:
            return compute_depth_in_subtree(individual, primitive, child_index) + 1
        else:
            child_index = child_subtree_slice.stop
    print("Not found!!!!")


def compute_depth_of_primitive(individual, primitive):
    return compute_depth_in_subtree(individual, primitive, 0)


def compute_depth_of_index(individual, index):
    return compute_depth_in_subtree(individual, individual[index], 0)


def evalBT(individual, toolbox=None):
    # individual, = mutUniform(individual, toolbox.pset())
    # print_individual(individual)
    # Transform the tree expression in a callable function
    toplevel = toolbox.compile(expr=individual)
    # toplevel = func()
    scores = bt.evaluate_controller(toplevel, toolbox.mission())
    return numpy.mean(scores),


def select_random_control_flow_node_index(individual):
    selected_node_index = None
    while selected_node_index is None:
        index = random.randrange(0, len(individual))
        if individual[index].ret is bt.ControlFlowNode:
            selected_node_index = index
    return selected_node_index

def mutNodeReplacement(individual, pset):
    if len(individual) < 2:
        return individual,

    index = select_random_control_flow_node_index(individual)
    node = individual[index]

    prims = [p for p in pset.primitives[bt.ControlFlowNode] if p.args == node.args]
    individual[index] = random.choice(prims)

    return individual,


def mutUniform(individual, pset):
    """
    Select a random node and replace it with another generated one.
    """
    # print_individual(individual)
    primitive = random.choice([i for i in individual if not isinstance(i, gp.Terminal)])
    depth = compute_depth_of_primitive(individual, primitive)
    index = individual.index(primitive)
    subtree_slice = individual.searchSubtree(index)  # the subtree defined by primitive
    new_subtree = generate_bt(pset, 1, 2, depth)
    individual[subtree_slice] = new_subtree
    # print_individual(individual)
    return individual,


def jones_mutation(individual, toolbox=None):
    prob_mut_u = 0.05  # uniform
    prob_mut_s = 0.1  # shrink
    prob_mut_n = 0.5  # node replacement
    prob_mut_e = 0.5  # ephemeral constant

    # print(individual[0].name)  # name is a string
    # print(individual[0].args)  # args is a list of classes (the classes for the arguments)
    # print(individual[0].arity)  # arity is an int
    # print(individual[0].ret)  # ret is the return type

    if random.random() <= prob_mut_u:
        mutUniform(individual, toolbox.pset())

    if random.random() <= prob_mut_s:
        gp.mutShrink(individual)

    if random.random() <= prob_mut_n:
        mutNodeReplacement(individual, toolbox.pset())

    if random.random() <= prob_mut_e:
        gp.mutEphemeral(individual, "one")

    return individual,

def elitist_ea(population, toolbox, nelite, cxpb, mutpb, ngen, stats=None,
               halloffame=None, verbose=__debug__):
    """This algorithm extends the eaSimple of deap (https://github.com/DEAP/deap/blob/master/deap/algorithms.py).
    See that algorithm for more documentation.
    This algorithm also introduces a new parameter nelite
    :param nelite: The number of elite individuals that are carried into the next generation without any modification.
    """
    logbook = tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if halloffame is not None:
        halloffame.update(population)

    record = stats.compile(population) if stats else {}
    logbook.record(gen=0, nevals=len(invalid_ind), **record)
    if verbose:
        print(logbook.stream)

    # Begin the generational process
    for gen in range(1, ngen + 1):
        # Create the elite set
        sorted_population = sorted(population, key=lambda x: x.fitness)
        elite = sorted_population[0:nelite]

        # Select the next generation individuals
        offspring = toolbox.select(population, len(population)-nelite)

        # Vary the pool of individuals
        offspring = algorithms.varAnd(offspring, toolbox, cxpb, mutpb)

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(offspring)

        # Replace the current population by the offspring
        population[:] = elite + offspring

        # Append the current generation statistics to the logbook
        record = stats.compile(population) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)
        if verbose:
            print(logbook.stream)

    return population, logbook

def genetic_programming(mission, budget, result_file):
    pset, toolbox, hof, stats = setup_gp()
    toolbox.register("select", tools.selTournament, tournsize=3)  # 1% of population size
    toolbox.register("mate", gp.cxOnePoint)
    toolbox.register("mutate", jones_mutation, toolbox=toolbox)
    toolbox.register("expr_mut", generate_bt, min_=1, max_=2)
    toolbox.register("pset", lambda: pset)
    toolbox.register("mission", lambda: mission)

    pop_size = 25
    number_of_elites = 3

    pop = toolbox.population(n=pop_size)

    prob_crossover = 0.8
    prob_mutation = 1  # always call the custom mutation function, the real mutation probabilities are set there
    generations = budget // (10 * pop_size)
    print("Running for {} generations".format(generations))

    elitist_ea(pop, toolbox, number_of_elites, prob_crossover, prob_mutation, generations, stats, halloffame=hof)

    best_individual = toolbox.compile(expr=hof[0]).to_cmd()

    print("Best controller found (Hall of Fame):")
    print(best_individual)

    with open(result_file, "w") as file:
        file.write(best_individual)

    return pop, stats, hof

if __name__ == "__main__":
    mission = sys.argv[1]
    budget = int(sys.argv[2])
    result_file = sys.argv[3]
    # genetic_programming("/home/jkuckling/AutoMoDe-missions/vanilla/for250sec/foraging_bt.argos", 1000)
    genetic_programming(mission, budget, result_file)
